// Project:		Generic
// File:		glmain.cpp
// Description:	This is a generic GL windows framework

#include <windows.h>
#include <GL/gl.h>
#include "usercode.h"

/////////////////  Constants & Variables ////////////////////////

#define FULLSCREEN 1		// Fullscreen flag: 1 = FULLSCREEN, 0 = window
#define WIDTH 640			// width of window/FULLSCREEN display
#define HEIGHT 480			// height of window/FULLSCREEN display

HDC hDC;					// dc store
HINSTANCE hInstance = 0;	// instance/module id
RECT wndRect;				// co-ords of corners of window on the screen
int num_modes = 0;			// number of available display modes
int current = -1;			// current display mode
BOOLEAN bQuit = FALSE;		// indicates program is quitting
DEVMODE devmode;			// current display mode being enumerated
DEVMODE *devmodes = NULL;	// pointer to array to store all devmodes
int fs_width;				// Window - width / Fullscreen - screen width
int fs_height;				// Window - Height / Fullscreen - screen height
int fs_bpp;					// bpp selected in changing display mode
int win_width, win_height;	// width/height of window

////////////////////  Display Mode Routines  //////////////////////

// Function to change display mode

BOOL ChangeMode(DWORD width, DWORD height, USHORT bpp)
{
	int i;					// counter
	int mode = -1;			// id of selected mode
	int success = FALSE;	// success flag

	// check pointer

	if (!devmodes)
		return FALSE;

	// find suitable mode

	for (i=0; i<num_modes; i++)
	{
		if (devmodes[i].dmPelsWidth == width &&
			devmodes[i].dmPelsHeight == height && 
			devmodes[i].dmBitsPerPel == bpp)
		{
			mode = i;
			break;
		}
	}

	// if suitable mode found, attempt to change display settings

	if (mode >= 0)
	{
		if (ChangeDisplaySettings(&devmodes[i], CDS_TEST) == DISP_CHANGE_SUCCESSFUL)
		{
			ChangeDisplaySettings(&devmodes[i], 0);
			current = i;
			success = TRUE;
		}
	}

	return success;
}

// Function to set/restore default display mode

void UseCurrentMode()
{
	ChangeDisplaySettings(NULL, 0);
}

//////////////////////////  Windows Functions  //////////////////////

// Function to handle windows messages

LONG WINAPI WindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{ 
	static PAINTSTRUCT ps;

	switch(uMsg)
	{
		case WM_PAINT:
			BeginPaint(hWnd, &ps);
			EndPaint(hWnd, &ps);
			return 0;

		case WM_SIZE:
			win_width = LOWORD(lParam);
			win_height = HIWORD(lParam);
			return 0;

		case WM_CHAR:
			switch (wParam)
			{
				case 27:			// ESC key
				bQuit = TRUE;
				break;
			}
			return 0;

		case WM_CLOSE:
			bQuit = TRUE;
			return 0;
	}

	return DefWindowProc(hWnd, uMsg, wParam, lParam); 
} 

// Function to create window

HWND CreateOpenGLWindow(char* title, int x, int y, int width, int height)
{
	HWND hWnd;		// window handle
	WNDCLASS wc;	// window class structure

	// register the window class

	if (!hInstance)
	{
		hInstance        = GetModuleHandle(NULL);
		wc.style         = CS_OWNDC;
		wc.lpfnWndProc   = (WNDPROC)WindowProc;
		wc.cbClsExtra    = 0;
		wc.cbWndExtra    = 0;
		wc.hInstance     = hInstance;
		wc.hIcon         = LoadIcon(NULL, IDI_WINLOGO);
		wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
		wc.hbrBackground = NULL;
		wc.lpszMenuName  = NULL;
		wc.lpszClassName = "OpenGL";

		if (!RegisterClass(&wc))
		{
			MessageBox(NULL, "RegisterClass() failed:  "
					   "Cannot register window class.", "Error",
					   MB_OK);
			return NULL;
		}
	}

	// create the window

	if (FULLSCREEN)
	{
		hWnd = CreateWindow("OpenGL", title, WS_POPUP | WS_MAXIMIZE |
							WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
							x, y, width, height, NULL, NULL, hInstance, NULL);
	}
	else
	{
		hWnd = CreateWindow("OpenGL", title, WS_OVERLAPPED | WS_SYSMENU |
							WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
							x, y, width, height, NULL, NULL, hInstance, NULL);
	}

	// handle window creation failure

	if (hWnd == NULL)
	{
		MessageBox(NULL, "CreateWindow() failed:  Cannot create a window.",
				   "Error", MB_OK);
		return NULL;
	}

	return hWnd;
}

// Function to select and set pixel format

BOOL SelectPixelFormat()
{
	int pf;						// pixel format index
	PIXELFORMATDESCRIPTOR pfd;	// pixel format descriptor

	// set up pixel format descriptor

	memset(&pfd, 0, sizeof(pfd));
	pfd.nSize        = sizeof(pfd);
	pfd.nVersion     = 1;
	pfd.dwFlags      = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
	pfd.iPixelType   = PFD_TYPE_RGBA;
	pfd.cColorBits   = fs_bpp;
	pfd.cDepthBits	 = fs_bpp;

	// choose a suitable pixel format

	pf = ChoosePixelFormat(hDC, &pfd);
	if (pf == 0)
	{
		MessageBox(NULL, "ChoosePixelFormat() failed:  "
				   "Cannot find a suitable pixel format.", "Error",
				   MB_OK); 
		return FALSE;
	} 
 
	// set chosen pixel format

	if (SetPixelFormat(hDC, pf, &pfd) == FALSE)
	{
		MessageBox(NULL, "SetPixelFormat() failed:  "
				   "Cannot set format specified.", "Error", MB_OK);
		return FALSE;
	}

	return TRUE;
}    

// Winmain

int APIENTRY WinMain(HINSTANCE hCurrentInst, HINSTANCE hPreviousInst,
					 LPSTR lpszCmdLine, int nCmdShow)
{
	HGLRC hRC;					// opengl context
	HWND hWnd;					// window handle
	MSG msg;					// message
	HACCEL  accel;				// accelerator table handle
	int win_caption;			// size of window title bar
	int win_pos_x;				// x position of upper left corner of window
	int win_pos_y;				// y position of upper left corner of window
	BOOL change_settings;		// display settings need changing

	// initialize variables

	fs_width = WIDTH;
	fs_height = HEIGHT;

	// create window

	if (FULLSCREEN)
	{
		win_width = GetSystemMetrics(SM_CXSCREEN);
		win_height = GetSystemMetrics(SM_CYSCREEN);
		if (win_width == fs_width && win_height == fs_height)
			change_settings = FALSE;
		else
			change_settings = TRUE;
		hWnd = CreateOpenGLWindow("glmain", 0, 0, win_width, win_height);
	}
	else
	{
		win_width = WIDTH;
		win_height = HEIGHT;
		win_caption = GetSystemMetrics(SM_CYCAPTION);
		if (win_width > (GetSystemMetrics(SM_CXSCREEN) - 6))
		{
			win_width = GetSystemMetrics(SM_CXSCREEN) - 6;
			win_height = GetSystemMetrics(SM_CYSCREEN) - win_caption - 6;
		}
		win_pos_x = (GetSystemMetrics(SM_CXSCREEN) / 2) - ((win_width + 6) / 2);
		win_pos_y = (GetSystemMetrics(SM_CYSCREEN) / 2) - ((win_height + win_caption + 6) / 2);
		hWnd = CreateOpenGLWindow("glmain", win_pos_x, win_pos_y, (win_width + 6), (win_height + win_caption + 6));
	}
	if (hWnd == NULL)
		exit(1);

	// enumerate display modes if required

	if(change_settings)
	{
		// get the total number of modes so we can allocate memory

		num_modes = 0;
		while (EnumDisplaySettings(NULL, num_modes, &devmode))
		num_modes++;

		// fill an array with all the devmodes

		devmodes = new DEVMODE[num_modes];
		if (devmodes)
		{
			num_modes = 0;
			while (EnumDisplaySettings(NULL, num_modes, &devmodes[num_modes]))
				num_modes++;
		}
	}

	// change to the display mode we want

	hDC = GetDC(hWnd);
	fs_bpp = GetDeviceCaps(hDC, BITSPIXEL);
	if (fs_bpp < 16)
	{
		MessageBox(hWnd, "Please change your desktop colour palette to 16bit or higher",
				   "RGB palette required", MB_OK);
		ReleaseDC(hWnd, hDC);
		DestroyWindow(hWnd);
		exit(1);
	}
	if (FULLSCREEN && change_settings)
	{
		if (!ChangeMode(fs_width, fs_height, fs_bpp))
		{
			fs_width = GetSystemMetrics(SM_CXSCREEN);
			fs_height = GetSystemMetrics(SM_CYSCREEN);
			MessageBox(hWnd, "Failed to set the selected display resolution" ,"Warning", MB_OK);
		}
		else
		{
			MoveWindow(hWnd, 0, 0, fs_width, fs_height, FALSE);		// win98 fix
		}
		if (devmodes)
		{
			delete devmodes;
			devmodes = NULL;
		}
	}
	if (FULLSCREEN)
	{
		ShowCursor(FALSE);
	}

	// set up GL & window

	if (!SelectPixelFormat())
	{
		if (FULLSCREEN)
		{
			ChangeDisplaySettings(NULL, 0);
			ShowCursor(TRUE);
		}
		ReleaseDC(hWnd, hDC);
		DestroyWindow(hWnd);
		exit(1);
	}
	hRC = wglCreateContext(hDC);
	wglMakeCurrent(hDC, hRC);
	ShowWindow(hWnd, SW_SHOW);

	// application specific initialisation

	init();

	// message loop

	accel = LoadAccelerators(hCurrentInst, "AppAccel");

	while (!bQuit)
	{	
		while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{
			if (!TranslateAccelerator(msg.hwnd, accel, &msg))
			{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}

		// calls to application update functions
	
		update();
		redraw();
	}

	// clean up

	wglMakeCurrent(NULL, NULL);
	wglDeleteContext(hRC);
	if (FULLSCREEN)
		ShowCursor(TRUE);
	if (FULLSCREEN && change_settings)
		ChangeDisplaySettings(NULL, 0);
	ReleaseDC(hWnd, hDC);
	DestroyWindow(hWnd);

	return 0;
}

